
//----------------------------------------------------------------
// cloud of probability (of robber existance)
//----------------------------------------------------------------

var pcloud

function pcloud_robber_move()
{
	if( "rob" in world ) // visible
	{
		pcloud = {}
		pcloud[world.rob.pos] = 1.0
		return
	}

	if( pcloud === undefined ) // init
	{
		pcloud = {}
		pcloud[world.rb] = 1.0
	}

	// move
	var n2p = {}
	for( var n in pcloud )
	{
		var pr = pcloud[n] / (1+world.graph[n].foot.length)
			{ if( !(n in n2p) ) n2p[n]=0 ; n2p[n]+=pr*(1.1-z(pcloud[n]))}
		for each( var e in world.graph[n].foot )
			{ if( !(e in n2p) ) n2p[e]=0 ; n2p[e]+=pr*(1.1-z(pcloud[e]))}
	}
	for each(var cp in world.cop)
		delete n2p[cp.pos]
	pcloud = n2p
	pcloud_normalize()
}

function pcloud_normalize()
{
	var sum = 0.0
	for( var n in pcloud )
		sum += pcloud[n];

	if( !sum || sum<=0.0 ) { // error recovery
		var p = 1 / sizeof(world.bank)
		for( var n in world.bank )
			pcloud[n] = p
	}
	else // normalize
		for( var n in pcloud )
			pcloud[n] /= sum;
}

function pcloud_smell(s)
{
	for( var n in pcloud )
		if( !(n in s) )
			delete pcloud[n]
	pcloud_normalize()
}

function pcloud_maybe_smell(s, cer)
{
	for( var n in pcloud )
		if( !(n in s) && n in pcloud )
			pcloud[n] *= (100-cer) / 100
	pcloud_normalize()
}

function pcloud_nosmell(s)
{
	for( var n in s )
		delete pcloud[n]
	pcloud_normalize()
}

function pcloud_ev(ev)
{
	var s = closure(ev.loc, (world.serial-ev.ser+1)/2, "foot", false)
	pcloud_smell(s)
}

//------------------------------------------------------------
// set of nodes reachable from src
//  in (exactly) n step by veh(foot|car)
//------------------------------------------------------------

function closure(src, n, veh, exact)
{
	var nc = {}
	if( n == 0 )
		nc[""+src] = 1;
	else
	{
		var nex = closure(src, n-1, veh, false)
		for( var pt in nex )
		{
			if( !exact )
				nc[pt] = 1
			for each(var adj in world.graph[pt][veh])
				if( !exact || !(adj in nex) )
			        nc[adj] = 1
		}
    }
	return nc
}

//------------------------------------------------------------
// normal pcloud updates and info sending
//------------------------------------------------------------

function on_info()
{
	pcloud_robber_move()

	var inf = []
	if( "smell" in world )
	{
		var cand = closure(self.pos, world.smell, self.veh(), true)
		var cert = Math.floor(100 / sizeof(cand))
		for(var c in cand)
			inf.push( [world.name.robber,c,"robber",world.serial,cert] )
		pcloud_smell(cand)
	}
	else
	{
		var step = (self.veh()=="foot" ? 2 : 1)
		var cand = closure(self.pos, step, self.veh(), false)
		for(var c in cand)
			inf.push( [world.name.robber,c,"robber",world.serial,-100] )
		pcloud_nosmell(cand)
	}
	for each( var ev in world.ev )
	{
		inf.push( [world.name.robber,ev.loc,"robber",ev.ser,100] )
		pcloud_ev(ev)
	}
	return inf
}

//------------------------------------------------------------
// processing informations from others
//------------------------------------------------------------

function on_info_others()
{
	var nonexist={}
	for each(var inf in info_msg) {
		var exist={}, cerSum=0

		// per person
		for each(var i in inf) {
			var bot = i[0]
			var loc = i[1]
			var pt  = i[2]
			var ser = Number(i[3])
			var cer = Number(i[4])
			if( world.serial-1 <= ser )
			{
				if( cer == -100 ) nonexist[loc]=1
				if( cer > 0  )       exist[loc]=1, cerSum+=cer
			}
		}

		// 95 < 100/6*6, 100/8*8, etc
		if( cerSum>95 ) pcloud_smell( exist )
		else            pcloud_maybe_smell( exist, cerSum )
	}
	pcloud_nosmell( nonexist )
}

//------------------------------------------------------------
// plannings
//------------------------------------------------------------

var myplan
function on_plan()
{
	on_info_others()

	var possibleNodes = []
	for( var n in pcloud )
		possibleNodes.push(n)
	possibleNodes.sort( function(x,y){
		return pcloud[y]-pcloud[x] })

	return myplan = on_my_plan( pcloud, possibleNodes )
}

function my_move_of_the_plan(p)
{
	var nearMe = {foot:{}, car:{}}
	nearMe.foot[self.pos] = 1
	nearMe.car [self.pos] = 1
	for each(var pt in world.graph[self.pos].foot) nearMe.foot[pt]= 1
	for each(var pt in world.graph[self.pos].car)  nearMe.car[pt] = 1

	if(p)
	for each(var pl in p)
		if( pl[0]==self.name &&
		  (Number(pl[3])==world.serial+1 || Number(pl[3])==world.serial+2))
		{
			var pos=pl[1], ptype=pl[2]
			if( self.ptype!=ptype && self.pos!=world.hq )
				continue
			if( !(pos in nearMe[ptype=="cop-car"?"car":"foot"]) )
			    continue
			return {pos:pos, ptype:ptype}
		}
	return undefined
}

//------------------------------------------------------------
// voting
//------------------------------------------------------------

var voteWinnerHistory = []
function lastWinner_is(p) {
	for(var i=voteWinnerHistory.length-1; i>=0; --i)
	    if( voteWinnerHistory[i] )
           	return voteWinnerHistory[i] == p
	return false
}

var vote_me = false
function on_vote()
{
	// make sure ...
	plan_msg[self.name] = myplan
	for each(var cp in world.name.cop) {
		if( !(cp in plan_msg) ) plan_msg[cp] = {mymove: undefined}
	    else plan_msg[cp].mymove = my_move_of_the_plan(plan_msg[cp])
	}

	// evaluate
	var vws  = (vote_me ? [self.name] : [])
	var oks  = []
	var bads = []
	for each(var cp in world.name.cop)
	    if(!vote_me || cp!=self.name)
	      if( plan_msg[cp].mymove )
		    if(lastWinner_is(cp)) vws.push(cp); else oks.push(cp)
		  else
	        bads.push(cp)

	return vws.concat(oks, bads)
}

//------------------------------------------------------------
// move base on the vote
//------------------------------------------------------------

function on_move()
{
	voteWinnerHistory.push(vote_msg)
	var vw     = (vote_msg ? vote_msg : self.name)
	var vwPlan = plan_msg[vw]

	var mm = vwPlan.mymove
	if( !mm ) mm = plan_msg[self.name].mymove

	if( mm )
	{
	    self.pos   = mm.pos
		self.ptype = mm.ptype
	}
}
