loadClass("util")
loadClass("msg")
loadClass("algorithm")

//===========================================================================
// CoreSystem
//  - msg_in
//  - msg_out
//  - world {serial, bank, ev, smell, rob, cop}
//                 -> {graph, hq, rb, exchange}
//  - info_msg (name=>array of (bot loc ...))
//  - plan_msg (name=>array of (bot loc ...))
//  - vote_msg (undef | winner's name)
//  - self  {name, ptype, pos}
//===========================================================================

function CoreSystem( p_name, p_type )
{
	var msg_in  = new MessageReader()
	var msg_out = new MessageWriter()
	var static_world = {graph:{}, hq:null, rb:null, exchange:{}, name:{}}
	var self

	//
	// send
	//   <reg>name, type</reg>
	//
	msg_out.write( "reg", [p_name, p_type] )

	//
	// recv
	//   <wsk>
	//     name
	//     robber
	//     cop[]
	//     nod.nod[loc typ x y]
	//     edg.edg[src dst typ]
	//   </wsk>
	//
	with( msg_in.read() )
	{
		var G = static_world.graph
		for each(var n in wsk[0].nod[0].nod)
		{
			G[n[0]] = {}
			switch( n[1] )
			{
			case "hq":
				static_world.hq     = n[0]
				static_world.exchange[n[0]] = {foot:"car", car:"foot"}
				break
			case "robber-start":
				static_world.rb = n[0]
				break
			case "bank":
			case "ordinary":
				break
			}
		}

		for each(var e in wsk[0].edg[0].edg)
		{
			function add_vert(src, dst, typ)
			{
				if( typ in G[src] )
				{
					for each(var n in G[src][typ])
						if( n==dst )
							return
					G[src][typ].push(dst)
				}
				else
					G[src][typ] = [dst]
			}

			switch( e[2] )
			{
			case "car":
				add_vert( e[0], e[1], "car" )
				break
			case "foot":
				add_vert( e[0], e[1], "car" )
				add_vert( e[0], e[1], "foot" )
				add_vert( e[1], e[0], "foot" )
				break
			}
		}

		static_world.name.self   = wsk[0].name[0]
		static_world.name.robber = wsk[0].robber[0]
		if( p_type == "robber" )
			static_world.name.cop = wsk[0].cop
		else {
			static_world.name.cop = [wsk[0].name[0]]
			for each(var nm in wsk[0].cop)
				if( nm != wsk[0].name[0] )
					static_world.name.cop.push(nm)
		}
		self = new person( wsk[0].name[0], p_type,
		                   p_type=="robber"?static_world.rb:static_world.hq )
	}

	//
	// recv
	//   <game-over-msg>
	//   <wor>
	//     wor
	//     smell
	//     ev.ev[]
	//     pl.pl[]
	//     bv.bv[]
	//   </wor>
	//
	this.update_world = function()
	{
		var msg = msg_in.read()
		if( "game-over" in msg )
			quit()

		var world = new Prototyped(static_world)
		with(msg)
		{
			world.serial = Number(wor[0].wor[0])
			if( wor[0].smell[0] > 0 )
				world.smell = Number(wor[0].smell[0])

			world.ev = []
			for each(var ev in wor[0].ev[0].ev)
				world.ev.push( {loc:ev[0], ser:Number(ev[1])} )

			world.bank = {}
			for each(var bv in wor[0].bv[0].bv)
				world.bank[bv[0]] = bv[1]

			world.cop = []
			for each(var pl in wor[0].pl[0].pl)
				if( pl[2] == "robber" )
					world.rob = new person(pl[0], pl[2], pl[1])
				else
					world.cop.push( new person(pl[0], pl[2], pl[1]) )
		}
		this.world = world
	}

	//
	// send
	//   <mov>dest, ptype</mov>
	//
	this.send_move = function()
	{
		msg_out.write( "mov", [self.pos, self.ptype] )
	}

	//---------------------------------------------------
	// person object
	//---------------------------------------------------

	function person(n,pt,ps)
	{
		this.name  = n
		this.ptype = pt
		this.pos   = ps
	}
	person.prototype.job = function()
	{
		return this.ptype=="robber" ? "robber" : "cop"
	}
	person.prototype.veh = function()
	{
		return this.ptype=="cop-car" ? "car" : "foot"
	}

	//---------------------------------------------------
	// this fields
	//---------------------------------------------------

	this.msg_in  = msg_in
	this.msg_out = msg_out
	this.self    = self
	this.world   = new Prototyped(static_world)
}



//===========================================================================
// RobberSystem
//   Globaly
//     - msg_in
//     - msg_out
//   on_init
//   TurnInit
//     - world
//     - self
//   on_move <== should modify self.pos
//===========================================================================

function ROBBER( p_name )
{
	this.__proto__= new CoreSystem( p_name, "robber" )
	if(this.on_init) this.on_init()
	for(;;)
	{
		this.update_world()
		if(this.on_move) this.on_move()
		this.send_move()
	}
}



//===========================================================================
// CopSystem
//   Globaly
//     - msg_in
//     - msg_out
//   on_init
//   TurnInit
//     - world
//     - self
//   on_info <== should return <inf></inf>s array
//     - info
//   on_plan <== should return <plan></plan>s array
//     - plan
//   on_vote <== should return <vote></vote> array
//     - vote
//   on_move <== should modify self.pos
//===========================================================================

function COP( p_name, p_type )
{
	if( p_type === undefined )
		p_type = cointoss(0.4) ? "cop-car" : "cop-foot" 

	this.__proto__ = new CoreSystem( p_name, p_type )
	if(this.on_init) this.on_init()
	for(;;)
	{
		var w = update_world()

        // info-write
		{
			this.msg_out.begin( "inf" )
			if(this.on_info) {
				var v = this.on_info()
				if(v)
				    for(var i in v)
				        this.msg_out.write( "inf", v[i] )
			}
			this.msg_out.end()
		}

        // info-read
		{
			this.info_msg = {}
			with( msg_in.read().from[0] )
				for(var i=0; i<from.length; ++i)
					this.info_msg[from[i]] = inf[i].inf
		}

		// plan-write
		{
			this.msg_out.begin( "plan" )
			if(this.on_plan) {
				var v = this.on_plan()
				if(v)
				    for(var i in v)
					    this.msg_out.write( "plan", v[i] )
			}
			this.msg_out.end()
		}

		// plan-read
		{
			this.plan_msg = {}
			with( msg_in.read().from[0] )
				for(var i=0; i<from.length; ++i)
					this.plan_msg[from[i]] = (plan[i].plan?plan[i].plan:[])
		}

		// vote-write
		{
			this.msg_out.begin( "vote" )
			var v = (this.on_vote ? this.on_vote() : world.name.cop)
			for(var i in v)
				this.msg_out.write( "vote", [v[i]] )
			this.msg_out.end()
		}

		// vote-read
		{
			this.vote_msg = undefined
			var msg = msg_in.read()
			if( "winner" in msg )
				this.vote_msg = msg.winner[0]
		}

		// move-write
		{
			if(this.on_move) this.on_move()
			this.send_move()
		}
	}
}
