
//===========================================================================
// MessageReader
//  - read()
//
//   The constructor for MessageReader objects, which parse the S->C
//   messages from standard input and generate message objects.
//===========================================================================

function MessageReader()
{
	var stdin =
	  new java.io.BufferedReader(
	    new java.io.InputStreamReader( java.lang.System["in"] ))

	//
	// Read one line of standard input as an array of strings
	//

	function readLine()
	{
		var s = new java.util.Scanner( stdin.readLine() )
		var a = []
		while( s.hasNext() )
			a.push( ""+s.next() )
		return a
	}

	//
	// Read one unit of message, and adds as a field of obj
	//

	function addMemberTo(obj)
	{
		var arr = readLine()
		var key = arr[0]
		var val = arr.slice(1)

		switch( key.slice(-1) )
		{
		case "/":
			return undefined
		case "\\":
			key = key.slice(0,-1)
			val = {}
			while( addMemberTo(val) );
			break
		case ":":
			key = key.slice(0,-1)
			val = (val.length==1 ? val[0] : val)
			break
		}

		if( key in obj ) obj[key].push( val )
		else             obj[key] = [val]
		return obj
	}


	//----------------------------------------------------------
	// read()
	//----------------------------------------------------------
	//  reads an object from message
	//    |aaa\
	//    |bbb: x y z
	//    |bbb: w u v
	//    |ccc
	//    |ddd\
	//    |eee: 123
	//    |ddd/
	//    |aaa/
	//  ===>
	//    aaa[0].bbb[0] = [x,y,z]
	//    aaa[0].bbb[1] = [w,u,v]
	//    aaa[0].ccc[0] = undefined
	//    aaa[0].ddd[0].eee[0] = 123
	//
	//ToDo:
	//  maybe we need to add some kind of schema supports in
	//  order to remove cumbersome [0]s ...
	//----------------------------------------------------------

	this.read = function()
	{
		return addMemberTo({})
	}
}



//===========================================================================
// MessageWriter
//  - begin(tag)
//  - end()
//  - write(tag, params)
//
//   The constructor for MessageReader objects, which generater the C->S
//   messages to standard output.
//ToDo:
//   automatic end-tag insertion via finally? (not so a good idea?)
//   anyway, current interface is too low level...
//===========================================================================

function MessageWriter()
{
	var stdout = java.lang.System.out
	var stack  = []

	//----------------------------------------------------------
	// begin(tag)
	//    prints "tag\"
	//----------------------------------------------------------

	this.begin = function(tag)
	{
		stdout.println( tag+"\\" )
		stack.push(tag)
	}

	//----------------------------------------------------------
	// end()
	//    prints "tag/"
	//----------------------------------------------------------

	this.end = function()
	{
		var tag = stack.pop()
		stdout.println( tag+"/" )
	}

	//----------------------------------------------------------
	// write(tag, params)
	//----------------------------------------------------------

	this.write = function(tag, params)
	{
		if( params === undefined )
			stdout.println( tag )
		else
			stdout.println( tag+": "+params.join(" ") )
	}
}
